use args::{Args, Command, CreateEmptyConfigArgs, GenerateScriptsArgs, InstallFromConfigArgs};
use clap::Parser;
use rusty_value::into_json::{EnumRepr, IntoJson, IntoJsonOptions};
use tokio::fs;
use tourmaline::{distro::OSConfig, error::AppResult, generate_script_files};

mod args;

#[tokio::main(flavor = "current_thread")]
async fn main() -> color_eyre::Result<()> {
    color_eyre::install().unwrap();
    dotenv::dotenv().unwrap();
    let args = Args::parse();

    match args.command {
        Command::InstallFromConfig(args) => install_from_config(args).await,
        Command::GenerateScripts(args) => generate_scripts(args).await,
        Command::CreateEmptyConfig(args) => generate_empty_config(args).await,
    }?;

    Ok(())
}

/// Installs the distro from a given configuration file
async fn install_from_config(args: InstallFromConfigArgs) -> AppResult<()> {
    tourmaline::create_executor(args.path)
        .await?
        .with_base_tasks()
        .with_custom_tasks()
        .execute()
        .await
}

async fn generate_scripts(args: GenerateScriptsArgs) -> AppResult<()> {
    generate_script_files(args.path).await
}

async fn generate_empty_config(args: CreateEmptyConfigArgs) -> AppResult<()> {
    let config = OSConfig::empty().into_json_with_options(&IntoJsonOptions {
        enum_repr: EnumRepr::Untagged,
    });
    let config_string = serde_json::to_string_pretty(&config)?;
    fs::write(args.path, config_string).await?;

    Ok(())
}
